#include <stdio.h>
#include <math.h>
#include <stdlib.h>

#include "type-def.h"
#include "util.h"
#include "kernel.h"
#include "imageio.h"
#include "smoothing.h"
#include "para-util.h"
#include "hist-util.h"
#include "filter-util.h"
#include "special_bessel.h"

int main(int argc, char **argv)
{
  char paraFile[256];
  IMAGE_DATABASE imageDataBase;
  int windex, upindex;
  int wsize;
  char prefix[256];
  char fname[256];
  int i,k;
  CONTROL_INFO  para_setting;
  FILTERBANK  fbank;
  IMAGEINT   obs_img, syn_img;
  float  total_err, max_err, weighted_err;
  FILE *errfp;

  obs_img.nrow = 0;
  syn_img.nrow = 0;
  if (argc < 1) {
    printf("Please specify the parameter file: ");
    scanf("%s", paraFile);
  }
  else {
    strcpy(paraFile, argv[1]);
  }
  
  LoadParameter(paraFile, &para_setting);
  DisplayParameter(&para_setting);
  
  srand(para_setting.seed);

  fbank.nfilter = 0;

  imageDataBase.nimages =0;
  imageDataBase.num_ref_model =0;

  Read_Filters(para_setting.filterFile, &fbank);
  
  
  if (Load_Database(para_setting.databaseFile, &fbank,
		    &imageDataBase) != 0) {
    Create_Database(&para_setting, &fbank,
		    &imageDataBase);
    Save_Database(para_setting.databaseFile, &fbank,
		  &imageDataBase);
  }
  
  Initialize_Obs_Analytical(&fbank, &obs_img,
			    &para_setting);
  sprintf(fname,"%s_obs.pgm", para_setting.prefix);
  Write_An_Image_Int(&obs_img, (int)(256/para_setting.greyLevel),
		     fname);

  Initialize_Syn(&fbank, &syn_img, &para_setting);

  Scale_Target_Hist(&fbank, (float)(fbank.norm));

  sprintf(fname,"%s_syn_init.pgm", para_setting.prefix);
  Write_An_Image_Int(&syn_img, (int)(256/para_setting.greyLevel),
		     fname);
  srand(814);
  for (i=0; i < para_setting.learningIter; i++) {
    
    if ( (i%para_setting.learningUpdate) ==0) {
      Update_Syn_Filters(&fbank, &syn_img);
      Generate_Hist_From_Map(&fbank);
      printf("Updating lambdas ... \n");
      sprintf(fname,"%s_learn_Histogram.dat", para_setting.prefix);
      total_err = Update_Lambda(i,fname, 
		    &fbank,
		    &para_setting,
		    &max_err,
		    &weighted_err);
      Save_Lambda_To_File(&fbank,
			  para_setting.lambdaFile);
      sprintf(fname,"%s_Err.dat", para_setting.prefix);
      if (i==0) {
	errfp = fopen(fname,"wb");
      }
      else {
	errfp = fopen(fname,"ab");
      }
      fprintf(errfp,"%d with max_error = %8.6f ", i, max_err);
      fprintf(errfp,"weighted error: ");
      fprintf(errfp,"%8.6f ", weighted_err);
      fprintf(errfp,"Total = %8.6f \n", total_err);
      fclose(errfp);
      
      if (max_err < para_setting.learningMaxErrEpsilon) break;
    }
    if ( (i%para_setting.saveSynIter) ==0) {
      sprintf(fname,"%s_learning_syn_%d.pgm", para_setting.prefix,i);
      Write_An_Image_Int(&syn_img, (int)(256/para_setting.greyLevel),
			 fname);
    }
    Flip(&syn_img, &fbank, &para_setting);
  }
  sprintf(fname,"%s_learning_final_%d.pgm", para_setting.prefix,i);
  Write_An_Image_Int(&syn_img, (int)(256/para_setting.greyLevel),
			 fname);
  
  Free_Image_Int(&obs_img);
  Free_Image_Int(&syn_img);
  Free_Filter_Bank(&fbank);
  Free_Database(&imageDataBase);
  return 0;
}














